@@ -46,7 +46,8 @@ void MaximaKeywords::loadKeywords()
     m_variables = definition.keywordList(QLatin1String("MaximaVariable"));
 
     // This is missing in KSyntaxHighlighting.
-    // TODO: send upstream pull request with this
+    // https://phabricator.kde.org/D18714
+    // OUTOFDATE: Remove after 5.55 KSyntaxHighlighting version
     m_variables << QLatin1String("%pi") << QLatin1String("%e") << QLatin1String(" %i")
                 << QLatin1String("%gamma") << QLatin1String("ind") << QLatin1String("inf")
                 << QLatin1String("infinity") << QLatin1String("minf") << QLatin1String("%phi")
@@ -55,9 +56,9 @@ void MaximaKeywords::loadKeywords()
     m_functions << QLatin1String("celine");
 
     // We use qBinarySearch with this lists
-    qSort(m_keywords);
-    qSort(m_functions);
-    qSort(m_variables);
+    std::sort(m_keywords.begin(), m_keywords.end());
+    std::sort(m_functions.begin(), m_functions.end());
+    std::sort(m_variables.begin(), m_variables.end());
 }
 
 const QStringList& MaximaKeywords::variables() const
