@@ -80,13 +80,24 @@ CantorShell::~CantorShell()
 {
     if (m_recentProjectsAction)
         m_recentProjectsAction->saveEntries(KSharedConfig::openConfig()->group(QLatin1String("Recent Files")));
+
+    if (!m_newBackendActions.isEmpty())
+    {
+        unplugActionList(QLatin1String("new_worksheet_with_backend_list"));
+        qDeleteAll(m_newBackendActions);
+        m_newBackendActions.clear();
+    }
+
+    unplugActionList(QLatin1String("view_show_panel_list"));
+    qDeleteAll(m_panels);
+    m_panels.clear();
 }
 
 void CantorShell::load(const QUrl &url)
 {
     if (!m_part||!m_part->url().isEmpty() || m_part->isModified() )
     {
-        addWorksheet(QLatin1String("null"));
+        addWorksheet(QString());
         m_tabWidget->setCurrentIndex(m_parts.size()-1);
     }
     if (!m_part->openUrl( url ))
@@ -95,18 +106,6 @@ void CantorShell::load(const QUrl &url)
         m_recentProjectsAction->addUrl(url);
 }
 
-bool CantorShell::hasAvailableBackend()
-{
-    bool hasBackend=false;
-    foreach(Cantor::Backend* b, Cantor::Backend::availableBackends())
-    {
-        if(b->isEnabled())
-            hasBackend=true;
-    }
-
-    return hasBackend;
-}
-
 void CantorShell::setupActions()
 {
     QAction* openNew = KStandardAction::openNew(this, SLOT(fileNew()), actionCollection());
@@ -179,18 +178,22 @@ void CantorShell::readProperties(const KConfigGroup & /*config*/)
     // in 'saveProperties'
 }
 
+/*!
+ * called when one of the "new backend" action or the "New" action are called
+ * adds a new worksheet with the backend assossiated with the called action
+ * or opens the "Choose Backend" dialog, respectively.
+ */
 void CantorShell::fileNew()
 {
-    if (sender()->inherits("QAction"))
+    QAction* a = static_cast<QAction*>(sender());
+    const QString& backendName = a->data().toString();
+    if (!backendName.isEmpty())
     {
-        QAction * a = qobject_cast<QAction*>(sender());
-        QString backendName = a->data().toString();
-        if (!backendName.isEmpty())
-        {
-            addWorksheet(backendName);
-            return;
-        }
+        addWorksheet(backendName);
+        return;
     }
+
+    //"New" action was called -> open the "Choose Backend" dialog.
     addWorksheet();
 }
 
@@ -208,12 +211,18 @@ void CantorShell::fileOpen()
     // this slot is called whenever the File->Open menu is selected,
     // the Open shortcut is pressed (usually CTRL+O) or the Open toolbar
     // button is clicked
-    QUrl url = QFileDialog::getOpenFileUrl(this, i18n("Open file"), QUrl(), i18n("Cantor Worksheet (*.cws)"));
+    static const QString& worksheetFilter = i18n("Cantor Worksheet (*.cws)");
+    static const QString& notebookFilter = i18n("Jupyter Notebook (*.ipynb)");
+    QString filter;
+    if (m_previousFilter == notebookFilter)
+        filter = notebookFilter + QLatin1String(";;") + worksheetFilter;
+    else
+        filter = worksheetFilter + QLatin1String(";;") + notebookFilter;
+    QUrl url = QFileDialog::getOpenFileUrl(this, i18n("Open file"), QUrl(), filter, &m_previousFilter);
 
     if (url.isEmpty() == false)
     {
-        // About this function, the style guide (
-        // http://developer.kde.org/documentation/standards/kde/style/basics/index.html )
+        // About this function, the style guide
         // says that it should open a new window if the document is _not_
         // in its initial state.  This is what we do here..
         /*if ( m_part->url().isEmpty() && ! m_part->isModified() )
@@ -234,7 +243,17 @@ void CantorShell::fileOpen()
 
 void CantorShell::addWorksheet()
 {
-    if(hasAvailableBackend()) //There is no point in asking for the backend, if no one is available
+    bool hasBackend = false;
+    for (auto* b : Cantor::Backend::availableBackends())
+    {
+        if(b->isEnabled())
+        {
+            hasBackend = true;
+            break;
+        }
+    }
+
+    if(hasBackend) //There is no point in asking for the backend, if no one is available
     {
         QString backend = Settings::self()->defaultBackend();
         if (backend.isEmpty())
@@ -260,7 +279,7 @@ void CantorShell::addWorksheet()
         int backendListSize = 0;
         foreach(Cantor::Backend* b, Cantor::Backend::availableBackends())
         {
-            if(!b->requirementsFullfilled()) //It's disabled because of misssing dependencies, not because of some other reason(like eg. nullbackend)
+            if(!b->requirementsFullfilled()) //It's disabled because of missing dependencies, not because of some other reason(like eg. nullbackend)
             {
                 backendList+=QString::fromLatin1("<li>%1: <a href=\"%2\">%2</a></li>").arg(b->name(), b->url());
                 ++backendListSize;
@@ -297,47 +316,52 @@ void CantorShell::addWorksheet(const QString& backendName)
     // this routine will find and load our Part.  it finds the Part by
     // name which is a bad idea usually.. but it's alright in this
     // case since our Part is made for this Shell
-    KPluginFactory* factory = KPluginLoader(QLatin1String("cantorpart")).factory();
+    KPluginLoader loader(QLatin1String("cantorpart"));
+    KPluginFactory* factory = loader.factory();
     if (factory)
     {
-        Cantor::Backend* backend = Cantor::Backend::getBackend(backendName);
-
-        if (backend)
+        if (!backendName.isEmpty())
         {
-            if (backend->isEnabled() || backendName == QLatin1String("null"))
+            Cantor::Backend* backend = Cantor::Backend::getBackend(backendName);
+            if (!backend)
             {
-                // now that the Part is loaded, we cast it to a Part to get our hands on it
-                KParts::ReadWritePart* part = factory->create<KParts::ReadWritePart>(m_tabWidget, QVariantList()<<backendName);
-                if (part)
-                {
-                    connect(part, SIGNAL(setCaption(QString,QIcon)), this, SLOT(setTabCaption(QString,QIcon)));
-                    connect(part, SIGNAL(worksheetSave(QUrl)), this, SLOT(onWorksheetSave(QUrl)));
-                    m_parts.append(part);
-
-                    int tab = m_tabWidget->addTab(part->widget(), QIcon::fromTheme(backend->icon()), i18n("Session %1", sessionCount++));
-                    m_tabWidget->setCurrentIndex(tab);
-                    // Setting focus on worksheet view, because Qt clear focus of added widget inside addTab
-                    // This fix https://bugs.kde.org/show_bug.cgi?id=395976
-                    part->widget()->findChild<QGraphicsView*>()->setFocus();
-                }
-                else
-                {
-                    qDebug()<<"error creating part ";
-                }
+                KMessageBox::error(this, i18n("Backend %1 is not installed", backendName), i18n("Cantor"));
+                return;
             }
             else
             {
-                KMessageBox::error(this, i18n("%1 backend installed, but inactive. Please check installation and Cantor settings", backendName), i18n("Cantor"));
+                if (!backend->isEnabled())
+                {
+                    KMessageBox::error(this, i18n("%1 backend installed, but inactive. Please check installation and Cantor settings", backendName), i18n("Cantor"));
+                    return;
+                }
             }
         }
+
+        // now that the Part is loaded, we cast it to a Part to get our hands on it
+        KParts::ReadWritePart* part = factory->create<KParts::ReadWritePart>(m_tabWidget, QVariantList()<<backendName);
+        if (part)
+        {
+            connect(part, SIGNAL(setCaption(QString,QIcon)), this, SLOT(setTabCaption(QString,QIcon)));
+            connect(part, SIGNAL(worksheetSave(QUrl)), this, SLOT(onWorksheetSave(QUrl)));
+            m_parts.append(part);
+
+            int tab = m_tabWidget->addTab(part->widget(), i18n("Session %1", sessionCount++));
+            m_tabWidget->setCurrentIndex(tab);
+            // Setting focus on worksheet view, because Qt clear focus of added widget inside addTab
+            // This fix https://bugs.kde.org/show_bug.cgi?id=395976
+            part->widget()->findChild<QGraphicsView*>()->setFocus();
+        }
         else
-            KMessageBox::error(this, i18n("Backend %1 is not installed", backendName), i18n("Cantor"));
+        {
+            qDebug()<<"error creating part ";
+        }
     }
     else
     {
         // if we couldn't find our Part, we exit since the Shell by
         // itself can't do anything useful
-        KMessageBox::error(this, i18n("Could not find the Cantor Part."));
+        KMessageBox::error(this, i18n("Failed to find the Cantor Part with error %1", loader.errorString()));
         qApp->quit();
         // we return here, cause qApp->quit() only means "exit the
         // next time we enter the event loop...
@@ -384,8 +408,11 @@ void CantorShell::setTabCaption(const QString& caption, const QIcon& icon)
     if (caption.isEmpty()) return;
 
     KParts::ReadWritePart* part=dynamic_cast<KParts::ReadWritePart*>(sender());
-    m_tabWidget->setTabText(m_tabWidget->indexOf(part->widget()), caption);
-    m_tabWidget->setTabIcon(m_tabWidget->indexOf(part->widget()), icon);
+    if (part)
+    {
+        m_tabWidget->setTabText(m_tabWidget->indexOf(part->widget()), caption);
+        m_tabWidget->setTabIcon(m_tabWidget->indexOf(part->widget()), icon);
+    }
 }
 
 void CantorShell::closeTab(int index)
@@ -647,7 +674,9 @@ void CantorShell::updatePanel()
 void CantorShell::updateNewSubmenu()
 {
     unplugActionList(QLatin1String("new_worksheet_with_backend_list"));
-    QList<QAction*> newBackendActions;
+    qDeleteAll(m_newBackendActions);
+    m_newBackendActions.clear();
+
     foreach (Cantor::Backend* backend, Cantor::Backend::availableBackends())
     {
         if (!backend->isEnabled())
@@ -655,9 +684,9 @@ void CantorShell::updateNewSubmenu()
         QAction * action = new QAction(QIcon::fromTheme(backend->icon()), backend->name(), nullptr);
         action->setData(backend->name());
         connect(action, SIGNAL(triggered()), this, SLOT(fileNew()));
-        newBackendActions << action;
+        m_newBackendActions << action;
     }
-    plugActionList(QLatin1String("new_worksheet_with_backend_list"), newBackendActions);
+    plugActionList(QLatin1String("new_worksheet_with_backend_list"), m_newBackendActions);
 }
 
 Cantor::WorksheetAccessInterface* CantorShell::currentWorksheetAccessInterface()
