@@ -19,18 +19,16 @@
  */
 
 #include "backend.h"
-using namespace Cantor;
+#include "extension.h"
 
-#include <KServiceTypeTrader>
-#include <KService>
-#include <QDebug>
-#include <KXMLGUIFactory>
-#include <KPluginFactory>
-#include <KPluginMetaData>
 #include <QDir>
+#include <QRegularExpression>
 #include <QUrl>
 
-#include "extension.h"
+#include <KPluginMetaData>
+#include <KLocalizedString>
+
+using namespace Cantor;
 
 class Cantor::BackendPrivate
 {
@@ -39,14 +37,13 @@ class Cantor::BackendPrivate
     QString comment;
     QString icon;
     QString url;
-    bool enabled;
+    bool enabled{true};
 };
 
 Backend::Backend(QObject* parent, const QList<QVariant>& args) : QObject(parent),
                                                                 d(new BackendPrivate)
 {
     Q_UNUSED(args)
-    d->enabled=true;
 }
 
 Backend::~Backend()
@@ -74,36 +71,25 @@ QString Backend::icon() const
     return d->icon;
 }
 
-QString Backend::version() const
-{
-    return QLatin1String();
-}
-
 QString Backend::url() const
 {
     return d->url;
 }
 
-QUrl Backend::helpUrl() const
-{
-    return QUrl();
-}
-
 bool Backend::isEnabled() const
 {
-    return d->enabled&&requirementsFullfilled();
+    return d->enabled && requirementsFullfilled();
 }
 
 void Backend::setEnabled(bool enabled)
 {
-    d->enabled=enabled;
+    d->enabled = enabled;
 }
 
 QStringList Backend::listAvailableBackends()
 {
-    QList<Backend* > backends=availableBackends();
     QStringList l;
-    foreach(Backend* b, backends)
+    for (Backend* b : availableBackends())
     {
         if(b->isEnabled())
             l<<b->name();
@@ -123,20 +109,19 @@ QList<Backend*> Backend::availableBackends()
     }
 
     QStringList pluginDirs;
-    foreach(const QString &dir, QCoreApplication::libraryPaths()){
+    for (const QString& dir : QCoreApplication::libraryPaths()){
         pluginDirs << dir + QDir::separator() + QLatin1String("cantor/backends");
     }
 
     QPluginLoader loader;
-    foreach(const QString &dir, pluginDirs){
-
+    for (const QString &dir : pluginDirs){
         qDebug() << "dir: " << dir;
         QStringList plugins;
         QDir pluginDir = QDir(dir);
 
         plugins = pluginDir.entryList();
 
-        foreach (const QString &plugin, plugins){
+        for (const QString &plugin : plugins){
             if (plugin==QLatin1String(".") || plugin==QLatin1String(".."))
                 continue;
 
@@ -148,7 +133,7 @@ QList<Backend*> Backend::availableBackends()
             }
 
             KPluginFactory* factory = KPluginLoader(loader.fileName()).factory();
-            Backend* backend = factory->create<Backend>();
+            Backend* backend = factory->create<Backend>(QCoreApplication::instance());
 
             KPluginMetaData info(loader);
             backend->d->name=info.name();
@@ -164,8 +149,7 @@ QList<Backend*> Backend::availableBackends()
 
 Backend* Backend::getBackend(const QString& name)
 {
-    QList<Backend*> backends=availableBackends();
-    foreach(Backend* b, backends)
+    for (Backend* b : availableBackends())
     {
         if(b->name().toLower()==name.toLower() || b->id().toLower()==name.toLower())
             return b;
@@ -174,24 +158,12 @@ Backend* Backend::getBackend(const QString& name)
     return nullptr;
 }
 
-QWidget* Backend::settingsWidget(QWidget* parent) const
-{
-    Q_UNUSED(parent)
-        return nullptr;
-}
-
-KConfigSkeleton* Backend::config() const
-{
-    return nullptr;
-}
-
-
 QStringList Backend::extensions() const
 {
-    QList<Extension*> extensions=findChildren<Extension*>(QRegExp(QLatin1String(".*Extension")));
+    QList<Extension*> extensions = findChildren<Extension*>(QRegularExpression(QLatin1String(".*Extension")));
     QStringList names;
-    foreach(Extension* e, extensions)
-        names<<e->objectName();
+    for (Extension* e : extensions)
+        names << e->objectName();
     return names;
 }
 
@@ -200,8 +172,35 @@ Extension* Backend::extension(const QString& name) const
     return findChild<Extension*>(name);
 }
 
-bool Backend::requirementsFullfilled(QString* const reason) const
+bool Backend::checkExecutable(const QString& name, const QString& path, QString* reason)
 {
-    Q_UNUSED(reason);
+    if (path.isEmpty())
+    {
+        if (reason)
+            *reason = i18n("No path for the %1 executable specified. "
+                        "Please provide the correct path in the application settings and try again.",
+                        name);
+        return false;
+    }
+
+    QFileInfo info(path);
+    if (!info.exists())
+    {
+        if (reason)
+            *reason = i18n("The specified file '%1' for the %2 executable doesn't exist. "
+                        "Please provide the correct path in the application settings and try again.",
+                        path, name);
+        return false;
+    }
+
+    if (!info.isExecutable())
+    {
+        if (reason)
+            *reason = i18n("The specified file '%1' doesn't point to an executable. "
+                        "Please provide the correct path in the application settings and try again.",
+                        name);
+        return false;
+    }
+
     return true;
 }
