@@ -21,6 +21,8 @@
 
 #include "worksheet.h"
 
+#include <QtGlobal>
+#include <QApplication>
 #include <QBuffer>
 #include <QDebug>
 #include <QDrag>
@@ -28,12 +30,15 @@
 #include <QPrinter>
 #include <QTimer>
 #include <QXmlQuery>
+#include <QJsonArray>
+#include <QJsonDocument>
 
 #include <KMessageBox>
 #include <KActionCollection>
 #include <KFontAction>
 #include <KFontSizeAction>
 #include <KToggleAction>
+#include <KLocalizedString>
 
 #include "settings.h"
 #include "commandentry.h"
@@ -43,6 +48,7 @@
 #include "imageentry.h"
 #include "pagebreakentry.h"
 #include "placeholderentry.h"
+#include "lib/jupyterutils.h"
 #include "lib/backend.h"
 #include "lib/extension.h"
 #include "lib/helpresult.h"
@@ -57,10 +63,11 @@ const double Worksheet::TopMargin = 12;
 const double Worksheet::EntryCursorLength = 30;
 const double Worksheet::EntryCursorWidth = 2;
 
-Worksheet::Worksheet(Cantor::Backend* backend, QWidget* parent)
+Worksheet::Worksheet(Cantor::Backend* backend, QWidget* parent, bool useDeafultWorksheetParameters)
     : QGraphicsScene(parent)
 {
-    m_session = backend->createSession();
+    m_session = nullptr;
+
     m_highlighter = nullptr;
 
     m_firstEntry = nullptr;
@@ -68,13 +75,16 @@ Worksheet::Worksheet(Cantor::Backend* backend, QWidget* parent)
     m_lastFocusedTextItem = nullptr;
     m_dragEntry = nullptr;
     m_placeholderEntry = nullptr;
-    m_viewWidth = 0;
-    m_protrusion = 0;
     m_dragScrollTimer = nullptr;
 
     m_choosenCursorEntry = nullptr;
     m_isCursorEntryAfterLastEntry = false;
 
+    m_useDefaultWorksheetParameters = useDeafultWorksheetParameters;
+
+    m_viewWidth = 0;
+    m_maxWidth = 0;
+
     m_entryCursorItem = addLine(0,0,0,0);
     const QColor& color = (palette().color(QPalette::Base).lightness() < 128) ? Qt::white : Qt::black;
     QPen pen(color);
@@ -86,49 +96,49 @@ Worksheet::Worksheet(Cantor::Backend* backend, QWidget* parent)
     connect(m_cursorItemTimer, &QTimer::timeout, this, &Worksheet::animateEntryCursor);
     m_cursorItemTimer->start(500);
 
-    m_isPrinting = false;
-    m_loginDone = false;
-    m_readOnly = false;
-    m_isLoadingFromFile = false;
+    m_jupyterMetadata = nullptr;
 
-    enableHighlighting(Settings::self()->highlightDefault());
-    enableCompletion(Settings::self()->completionDefault());
-    enableExpressionNumbering(Settings::self()->expressionNumberingDefault());
-    enableAnimations(Settings::self()->animationDefault());
+    if (backend)
+        initSession(backend);
 }
 
 Worksheet::~Worksheet()
 {
-    // This is necessary, because a SeachBar might access firstEntry()
+    // This is necessary, because a SearchBar might access firstEntry()
     // while the scene is deleted. Maybe there is a better solution to
     // this problem, but I can't seem to find it.
     m_firstEntry = nullptr;
-    if (m_loginDone)
+    if (m_session && m_session->status() != Cantor::Session::Disable)
         m_session->logout();
     if (m_session)
     {
-        disconnect(m_session, 0, 0, 0);
+        disconnect(m_session, nullptr, nullptr, nullptr);
         if (m_session->status() != Cantor::Session::Disable)
             m_session->logout();
         m_session->deleteLater();
         m_session = nullptr;
     }
+    if (m_jupyterMetadata)
+        delete m_jupyterMetadata;
 }
 
 void Worksheet::loginToSession()
 {
     m_session->login();
 #ifdef WITH_EPS
-    session()->setTypesettingEnabled(Settings::self()->typesetDefault());
+    if (Cantor::LatexRenderer::isLatexAvailable())
+        session()->setTypesettingEnabled(Settings::self()->typesetDefault());
+    else
+        session()->setTypesettingEnabled(false);
 #else
     session()->setTypesettingEnabled(false);
 #endif
-    m_loginDone = true;
 }
 
 void Worksheet::print(QPrinter* printer)
 {
     m_epsRenderer.useHighResolution(true);
+    m_mathRenderer.useHighResolution(true);
     m_isPrinting = true;
     QRect pageRect = printer->pageRect();
     qreal scale = 1; // todo: find good scale for page size
@@ -166,6 +176,7 @@ void Worksheet::print(QPrinter* printer)
     painter.end();
     m_isPrinting = false;
     m_epsRenderer.useHighResolution(false);
+    m_mathRenderer.useHighResolution(false);
     m_epsRenderer.setScale(-1);  // force update in next call to setViewSize,
     worksheetView()->updateSceneSize(); // ... which happens in here
 }
@@ -182,6 +193,7 @@ void Worksheet::setViewSize(qreal w, qreal h, qreal s, bool forceUpdate)
     m_viewWidth = w;
     if (s != m_epsRenderer.scale() || forceUpdate) {
         m_epsRenderer.setScale(s);
+        m_mathRenderer.setScale(s);
         for (WorksheetEntry *entry = firstEntry(); entry; entry = entry->next())
             entry->updateEntry();
     }
@@ -202,7 +214,8 @@ void Worksheet::updateLayout()
     const qreal x = LeftMargin;
     for (WorksheetEntry *entry = firstEntry(); entry; entry = entry->next())
         y += entry->setGeometry(x, y, w);
-    setSceneRect(QRectF(0, 0, m_viewWidth + m_protrusion, y));
+
+    setSceneRect(QRectF(0, 0, sceneRect().width(), y));
     if (cursorRectVisible)
         makeVisible(worksheetCursor());
     else if (atEnd)
@@ -224,7 +237,7 @@ void Worksheet::updateEntrySize(WorksheetEntry* entry)
         entry->setY(y);
         y += entry->size().height();
     }
-    setSceneRect(QRectF(0, 0, m_viewWidth + m_protrusion, y));
+    setSceneRect(QRectF(0, 0, sceneRect().width(), y));
     if (cursorRectVisible)
         makeVisible(worksheetCursor());
     else if (atEnd)
@@ -232,39 +245,35 @@ void Worksheet::updateEntrySize(WorksheetEntry* entry)
     drawEntryCursor();
 }
 
-void Worksheet::addProtrusion(qreal width)
+void Worksheet::setRequestedWidth(QGraphicsObject* object, qreal width)
 {
-    if (m_itemProtrusions.contains(width))
-        ++m_itemProtrusions[width];
-    else
-        m_itemProtrusions.insert(width, 1);
-    if (width > m_protrusion) {
-        m_protrusion = width;
-        qreal y = lastEntry()->size().height() + lastEntry()->y();
-        setSceneRect(QRectF(0, 0, m_viewWidth + m_protrusion, y));
+    qreal oldWidth = m_itemWidths[object];
+    m_itemWidths[object] = width;
+
+    if (width > m_maxWidth || oldWidth == m_maxWidth)
+    {
+        m_maxWidth = width;
+        qreal y = lastEntry() ? lastEntry()->size().height() + lastEntry()->y() : 0;
+        setSceneRect(QRectF(0, 0, m_maxWidth + LeftMargin + RightMargin, y));
     }
 }
 
-void Worksheet::updateProtrusion(qreal oldWidth, qreal newWidth)
+void Worksheet::removeRequestedWidth(QGraphicsObject* object)
 {
-    removeProtrusion(oldWidth);
-    addProtrusion(newWidth);
-}
+    if (!m_itemWidths.contains(object))
+        return;
 
-void Worksheet::removeProtrusion(qreal width)
-{
-    if (--m_itemProtrusions[width] == 0) {
-        m_itemProtrusions.remove(width);
-        if (width == m_protrusion) {
-            qreal max = -1;
-            for (qreal p : m_itemProtrusions.keys()) {
-                if (p > max)
-                    max = p;
-            }
-            m_protrusion = max;
-            qreal y = lastEntry()->size().height() + lastEntry()->y();
-            setSceneRect(QRectF(0, 0, m_viewWidth + m_protrusion, y));
-        }
+    qreal width = m_itemWidths[object];
+    m_itemWidths.remove(object);
+
+    if (width == m_maxWidth)
+    {
+        m_maxWidth = 0;
+        for (qreal width : m_itemWidths.values())
+            if (width > m_maxWidth)
+                m_maxWidth = width;
+        qreal y = lastEntry() ? lastEntry()->size().height() + lastEntry()->y() : 0;
+        setSceneRect(QRectF(0, 0, m_maxWidth + LeftMargin + RightMargin, y));
     }
 }
 
@@ -309,7 +318,8 @@ WorksheetView* Worksheet::worksheetView()
 
 void Worksheet::setModified()
 {
-    emit modified();
+    if (!m_isLoadingFromFile)
+        emit modified();
 }
 
 WorksheetCursor Worksheet::worksheetCursor()
@@ -337,11 +347,11 @@ void Worksheet::setWorksheetCursor(const WorksheetCursor& cursor)
 
 WorksheetEntry* Worksheet::currentEntry()
 {
-    QGraphicsItem* item = focusItem();
     // Entry cursor activate
     if (m_choosenCursorEntry || m_isCursorEntryAfterLastEntry)
         return nullptr;
 
+    QGraphicsItem* item = focusItem();
     if (!item /*&& !hasFocus()*/)
         item = m_lastFocusedTextItem;
     /*else
@@ -491,17 +501,17 @@ void Worksheet::startDrag(WorksheetEntry* entry, QDrag* drag)
 void Worksheet::evaluate()
 {
     qDebug()<<"evaluate worksheet";
-    if (!m_loginDone && !m_readOnly)
+    if (!m_readOnly && m_session && m_session->status() == Cantor::Session::Disable)
         loginToSession();
 
     firstEntry()->evaluate(WorksheetEntry::EvaluateNext);
 
-    emit modified();
+    setModified();
 }
 
 void Worksheet::evaluateCurrentEntry()
 {
-    if (!m_loginDone && !m_readOnly)
+    if (!m_readOnly && m_session && m_session->status() == Cantor::Session::Disable)
         loginToSession();
 
     WorksheetEntry* entry = currentEntry();
@@ -535,13 +545,16 @@ WorksheetEntry* Worksheet::appendEntry(const int type, bool focus)
         if (!firstEntry())
             setFirstEntry(entry);
         setLastEntry(entry);
-        updateLayout();
-        if (focus)
+        if (!m_isLoadingFromFile)
         {
-            makeVisible(entry);
-            focusEntry(entry);
+            updateLayout();
+            if (focus)
+            {
+                makeVisible(entry);
+                focusEntry(entry);
+            }
+            setModified();
         }
-        emit modified();
     }
     return entry;
 }
@@ -614,7 +627,7 @@ WorksheetEntry* Worksheet::insertEntry(const int type, WorksheetEntry* current)
         else
             setLastEntry(entry);
         updateLayout();
-        emit modified();
+        setModified();
     } else {
         entry = next;
     }
@@ -686,7 +699,7 @@ WorksheetEntry* Worksheet::insertEntryBefore(int type, WorksheetEntry* current)
         else
             setFirstEntry(entry);
         updateLayout();
-        emit modified();
+        setModified();
     }
     else
         entry = prev;
@@ -745,14 +758,14 @@ void Worksheet::highlightItem(WorksheetTextItem* item)
         return;
 
     QTextDocument *oldDocument = m_highlighter->document();
-    QList<QList<QTextLayout::FormatRange> > formats;
+    QList<QVector<QTextLayout::FormatRange> > formats;
 
     if (oldDocument)
     {
         for (QTextBlock b = oldDocument->firstBlock();
              b.isValid(); b = b.next())
         {
-            formats.append(b.layout()->additionalFormats());
+            formats.append(b.layout()->formats());
         }
     }
 
@@ -772,7 +785,7 @@ void Worksheet::highlightItem(WorksheetTextItem* item)
         for (QTextBlock b = oldDocument->firstBlock();
              b.isValid(); b = b.next())
         {
-            b.layout()->setAdditionalFormats(formats.first());
+            b.layout()->setFormats(formats.first());
             formats.pop_front();
         }
         cursor.endEditBlock();
@@ -810,7 +823,7 @@ void Worksheet::rehighlight()
             for (QTextBlock b = item->document()->firstBlock();
                  b.isValid(); b = b.next())
             {
-                b.layout()->clearAdditionalFormats();
+                b.layout()->clearFormats();
             }
             cursor.endEditBlock();
         }
@@ -880,6 +893,16 @@ void Worksheet::enableAnimations(bool enable)
     m_animationsEnabled = enable;
 }
 
+bool Worksheet::embeddedMathEnabled()
+{
+    return m_embeddedMathEnabled && m_mathRenderer.mathRenderAvailable();
+}
+
+void Worksheet::enableEmbeddedMath(bool enable)
+{
+    m_embeddedMathEnabled = enable;
+}
+
 void Worksheet::enableExpressionNumbering(bool enable)
 {
     m_showExpressionIds=enable;
@@ -901,6 +924,40 @@ QDomDocument Worksheet::toXML(KZip* archive)
     return doc;
 }
 
+QJsonDocument Worksheet::toJupyterJson()
+{
+    QJsonDocument doc;
+    QJsonObject root;
+
+    QJsonObject metadata(m_jupyterMetadata ? *m_jupyterMetadata : QJsonObject());
+
+    QJsonObject kernalInfo;
+    if (m_session && m_session->backend())
+        kernalInfo = Cantor::JupyterUtils::getKernelspec(m_session->backend());
+    else
+        kernalInfo.insert(QLatin1String("name"), m_backendName);
+    metadata.insert(QLatin1String("kernelspec"), kernalInfo);
+
+    root.insert(QLatin1String("metadata"), metadata);
+
+    // Not sure, but it looks like we support nbformat version 4.5
+    root.insert(QLatin1String("nbformat"), 4);
+    root.insert(QLatin1String("nbformat_minor"), 5);
+
+    QJsonArray cells;
+    for( WorksheetEntry* entry = firstEntry(); entry; entry = entry->next())
+    {
+        const QJsonValue entryJson = entry->toJupyterJson();
+
+        if (!entryJson.isNull())
+            cells.append(entryJson);
+    }
+    root.insert(QLatin1String("cells"), cells);
+
+    doc.setObject(root);
+    return doc;
+}
+
 void Worksheet::save( const QString& filename )
 {
     QFile file(filename);
@@ -926,22 +983,40 @@ QByteArray Worksheet::saveToByteArray()
 void Worksheet::save( QIODevice* device)
 {
     qDebug()<<"saving to filename";
-    KZip zipFile( device );
-
-
-    if ( !zipFile.open(QIODevice::WriteOnly) )
+    switch (m_type)
     {
-        KMessageBox::error( worksheetView(),
-                            i18n( "Cannot write file." ),
-                            i18n( "Error - Cantor" ));
-        return;
+        case CantorWorksheet:
+        {
+            KZip zipFile( device );
+
+            if ( !zipFile.open(QIODevice::WriteOnly) )
+            {
+                KMessageBox::error( worksheetView(),
+                                    i18n( "Cannot write file." ),
+                                    i18n( "Error - Cantor" ));
+                return;
+            }
+
+            QByteArray content = toXML(&zipFile).toByteArray();
+            zipFile.writeFile( QLatin1String("content.xml"), content.data());
+            break;
+        }
+
+        case JupyterNotebook:
+        {
+            if (!device->isWritable())
+            {
+                KMessageBox::error( worksheetView(),
+                                    i18n( "Cannot write file." ),
+                                    i18n( "Error - Cantor" ));
+                return;
+            }
+
+            const QJsonDocument& doc = toJupyterJson();
+            device->write(doc.toJson(QJsonDocument::Indented));
+            break;
+        }
     }
-
-    QByteArray content = toXML(&zipFile).toByteArray();
-    qDebug()<<"content: "<<content;
-    zipFile.writeFile( QLatin1String("content.xml"), content.data());
-
-    /*zipFile.close();*/
 }
 
 
@@ -964,9 +1039,12 @@ void Worksheet::savePlain(const QString& filename)
         if (backend->extensions().contains(QLatin1String("ScriptExtension")))
         {
             Cantor::ScriptExtension* e=dynamic_cast<Cantor::ScriptExtension*>(backend->extension(QLatin1String(("ScriptExtension"))));
-            cmdSep=e->commandSeparator();
-            commentStartingSeq = e->commentStartingSequence();
-            commentEndingSeq = e->commentEndingSequence();
+            if (e)
+            {
+                cmdSep=e->commandSeparator();
+                commentStartingSeq = e->commentStartingSequence();
+                commentEndingSeq = e->commentEndingSequence();
+            }
         }
     }
     else
@@ -1019,9 +1097,10 @@ void Worksheet::saveLatex(const QString& filename)
 
 bool Worksheet::load(const QString& filename )
 {
+    qDebug() << "loading worksheet" << filename;
     QFile file(filename);
     if (!file.open(QIODevice::ReadOnly)) {
-        KMessageBox::error(worksheetView(), i18n("Couldn't open the file %1", filename), i18n("Cantor"));
+        KMessageBox::error(worksheetView(), i18n("Couldn't open the file %1.", filename), i18n("Open File"));
         return false;
     }
 
@@ -1041,39 +1120,64 @@ void Worksheet::load(QByteArray* data)
 
 bool Worksheet::load(QIODevice* device)
 {
-    KZip file(device);
-    if (!file.open(QIODevice::ReadOnly)) {
-        qDebug()<<"not a zip file";
+    if (!device->isReadable())
+    {
         QApplication::restoreOverrideCursor();
-        KMessageBox::error(worksheetView(), i18n("The selected file is not a valid Cantor project file."), i18n("Cantor"));
+        KMessageBox::error(worksheetView(), i18n("Couldn't open the selected file for reading."), i18n("Open File"));
         return false;
     }
 
-    const KArchiveEntry* contentEntry=file.directory()->entry(QLatin1String("content.xml"));
+    KZip archive(device);
+
+    if (archive.open(QIODevice::ReadOnly))
+        return loadCantorWorksheet(archive);
+    else
+    {
+        qDebug() <<"not a zip file";
+        // Go to begin of data, we need read all data in second time
+        device->seek(0);
+
+        QJsonParseError error;
+        const QJsonDocument& doc = QJsonDocument::fromJson(device->readAll(), &error);
+        if (error.error != QJsonParseError::NoError)
+        {
+            qDebug()<<"not a json file, parsing failed with error: " << error.errorString();
+            QApplication::restoreOverrideCursor();
+            KMessageBox::error(worksheetView(), i18n("The selected file is not a valid Cantor or Jupyter project file."), i18n("Open File"));
+            return false;
+        }
+        else
+            return loadJupyterNotebook(doc);
+    }
+}
+
+bool Worksheet::loadCantorWorksheet(const KZip& archive)
+{
+    m_type = Type::CantorWorksheet;
+
+    const KArchiveEntry* contentEntry=archive.directory()->entry(QLatin1String("content.xml"));
     if (!contentEntry->isFile())
     {
         qDebug()<<"content.xml file not found in the zip archive";
         QApplication::restoreOverrideCursor();
-        KMessageBox::error(worksheetView(), i18n("The selected file is not a valid Cantor project file."), i18n("Cantor"));
+        KMessageBox::error(worksheetView(), i18n("The selected file is not a valid Cantor project file."), i18n("Open File"));
         return false;
     }
 
     const KArchiveFile* content = static_cast<const KArchiveFile*>(contentEntry);
     QByteArray data = content->data();
 
-//     qDebug()<<"read: "<<data;
 
     QDomDocument doc;
     doc.setContent(data);
     QDomElement root=doc.documentElement();
-//     qDebug()<<root.tagName();
 
     m_backendName=root.attribute(QLatin1String("backend"));
     Cantor::Backend* b=Cantor::Backend::getBackend(m_backendName);
     if (!b)
     {
         QApplication::restoreOverrideCursor();
-        KMessageBox::information(worksheetView(), i18n("%1 backend was not found. Editing and executing entries is not possible", m_backendName), i18n("Cantor"));
+        KMessageBox::information(worksheetView(), i18n("%1 backend was not found. Editing and executing entries is not possible.", m_backendName), i18n("Open File"));
         m_readOnly = true;
     }
     else
@@ -1084,7 +1188,7 @@ bool Worksheet::load(QIODevice* device)
         QApplication::restoreOverrideCursor();
         KMessageBox::information(worksheetView(), i18n("There are some problems with the %1 backend,\n"\
                                             "please check your configuration or install the needed packages.\n"
-                                            "You will only be able to view this worksheet.", m_backendName), i18n("Cantor"));
+                                            "You will only be able to view this worksheet.", m_backendName), i18n("Open File"));
         m_readOnly = true;
     }
 
@@ -1100,9 +1204,8 @@ bool Worksheet::load(QIODevice* device)
     //cleanup the worksheet and all it contains
     delete m_session;
     m_session=nullptr;
-    m_loginDone = false;
 
-    //file can only be loaded in a worksheet that was not eidted/modified yet (s.a. CantorShell::load())
+    //file can only be loaded in a worksheet that was not edited/modified yet (s.a. CantorShell::load())
     //in this case on the default "first entry" is available -> delete it.
     if (m_firstEntry) {
         delete m_firstEntry;
@@ -1110,9 +1213,11 @@ bool Worksheet::load(QIODevice* device)
     }
 
     resetEntryCursor();
+    m_itemWidths.clear();
+    m_maxWidth = 0;
 
     if (!m_readOnly)
-        m_session=b->createSession();
+        initSession(b);
 
     qDebug()<<"loading entries";
     QDomElement expressionChild = root.firstChildElement();
@@ -1123,28 +1228,28 @@ bool Worksheet::load(QIODevice* device)
         if (tag == QLatin1String("Expression"))
         {
             entry = appendEntry(CommandEntry::Type, false);
-            entry->setContent(expressionChild, file);
+            entry->setContent(expressionChild, archive);
         } else if (tag == QLatin1String("Text"))
         {
             entry = appendEntry(TextEntry::Type, false);
-            entry->setContent(expressionChild, file);
+            entry->setContent(expressionChild, archive);
         } else if (tag == QLatin1String("Markdown"))
         {
             entry = appendEntry(MarkdownEntry::Type, false);
-            entry->setContent(expressionChild, file);
+            entry->setContent(expressionChild, archive);
         } else if (tag == QLatin1String("Latex"))
         {
             entry = appendEntry(LatexEntry::Type, false);
-            entry->setContent(expressionChild, file);
+            entry->setContent(expressionChild, archive);
         } else if (tag == QLatin1String("PageBreak"))
         {
             entry = appendEntry(PageBreakEntry::Type, false);
-            entry->setContent(expressionChild, file);
+            entry->setContent(expressionChild, archive);
         }
         else if (tag == QLatin1String("Image"))
         {
             entry = appendEntry(ImageEntry::Type, false);
-            entry->setContent(expressionChild, file);
+            entry->setContent(expressionChild, archive);
         }
 
         if (m_readOnly && entry)
@@ -1160,6 +1265,7 @@ bool Worksheet::load(QIODevice* device)
         clearFocus();
 
     m_isLoadingFromFile = false;
+    updateLayout();
 
     //Set the Highlighting, depending on the current state
     //If the session isn't logged in, use the default
@@ -1169,6 +1275,204 @@ bool Worksheet::load(QIODevice* device)
     return true;
 }
 
+void Worksheet::initSession(Cantor::Backend* backend)
+{
+    m_session = backend->createSession();
+    if (m_useDefaultWorksheetParameters)
+    {
+        enableHighlighting(Settings::self()->highlightDefault());
+        enableCompletion(Settings::self()->completionDefault());
+        enableExpressionNumbering(Settings::self()->expressionNumberingDefault());
+        enableAnimations(Settings::self()->animationDefault());
+        enableEmbeddedMath(Settings::self()->embeddedMathDefault());
+    }
+}
+
+bool Worksheet::loadJupyterNotebook(const QJsonDocument& doc)
+{
+    m_type = Type::JupyterNotebook;
+
+    int nbformatMajor, nbformatMinor;
+    if (!Cantor::JupyterUtils::isJupyterNotebook(doc))
+    {
+        // Two possibilities: old jupyter notebook (version <= 4.0.0 and a another scheme) or just not a notebook at all
+        std::tie(nbformatMajor, nbformatMinor) = Cantor::JupyterUtils::getNbformatVersion(doc.object());
+        if (nbformatMajor == 0 && nbformatMinor == 0)
+        {
+            QApplication::restoreOverrideCursor();
+            showInvalidNotebookSchemeError();
+        }
+        else
+        {
+            KMessageBox::error(worksheetView(),
+                i18n("Jupyter notebooks with versions lower than 4.5 (detected version %1.%2) are not supported.", nbformatMajor, nbformatMinor ),
+                i18n("Open File"));
+        }
+
+        return false;
+    }
+
+    QJsonObject notebookObject = doc.object();
+    std::tie(nbformatMajor, nbformatMinor) = Cantor::JupyterUtils::getNbformatVersion(notebookObject);
+
+    if (QT_VERSION_CHECK(nbformatMajor, nbformatMinor, 0) > QT_VERSION_CHECK(4,5,0))
+    {
+        QApplication::restoreOverrideCursor();
+        KMessageBox::error(
+            worksheetView(),
+            i18n("Jupyter notebooks with versions higher than 4.5 (detected version %1.%2) are not supported.", nbformatMajor, nbformatMinor),
+            i18n("Open File")
+        );
+        return false;
+    }
+
+    const QJsonArray& cells = Cantor::JupyterUtils::getCells(notebookObject);
+    const QJsonObject& metadata = Cantor::JupyterUtils::getMetadata(notebookObject);
+    if (m_jupyterMetadata)
+        delete m_jupyterMetadata;
+    m_jupyterMetadata = new QJsonObject(metadata);
+
+    const QJsonObject& kernalspec = metadata.value(QLatin1String("kernelspec")).toObject();
+    m_backendName = Cantor::JupyterUtils::getKernelName(kernalspec);
+    if (kernalspec.isEmpty() || m_backendName.isEmpty())
+    {
+        QApplication::restoreOverrideCursor();
+        showInvalidNotebookSchemeError();
+        return false;
+    }
+
+    Cantor::Backend* backend = Cantor::Backend::getBackend(m_backendName);
+    if (!backend)
+    {
+        QApplication::restoreOverrideCursor();
+        KMessageBox::information(worksheetView(),
+            i18n("%1 backend was not found. Editing and executing entries is not possible.", m_backendName),
+            i18n("Open File"));
+        m_readOnly = true;
+    }
+    else
+        m_readOnly = false;
+
+    if(!m_readOnly && !backend->isEnabled())
+    {
+        QApplication::restoreOverrideCursor();
+        KMessageBox::information(worksheetView(), i18n("There are some problems with the %1 backend,\n"\
+                                            "please check your configuration or install the needed packages.\n"
+                                            "You will only be able to view this worksheet.", m_backendName), i18n("Open File"));
+        m_readOnly = true;
+    }
+
+    if (m_readOnly)
+    {
+        for (QAction* action : m_richTextActionList)
+            action->setEnabled(false);
+    }
+
+
+    m_isLoadingFromFile = true;
+
+    if (m_session)
+        delete m_session;
+    m_session = nullptr;
+
+    if (m_firstEntry) {
+        delete m_firstEntry;
+        m_firstEntry = nullptr;
+    }
+
+    resetEntryCursor();
+    m_itemWidths.clear();
+    m_maxWidth = 0;
+
+    if (!m_readOnly)
+        initSession(backend);
+
+    qDebug() << "loading jupyter entries";
+
+    WorksheetEntry* entry = nullptr;
+    for (QJsonArray::const_iterator iter = cells.begin(); iter != cells.end(); iter++) {
+        if (!Cantor::JupyterUtils::isJupyterCell(*iter))
+        {
+            QApplication::restoreOverrideCursor();
+            QString explanation;
+            if (iter->isObject())
+                explanation = i18n("an object with keys: %1", iter->toObject().keys().join(QLatin1String(", ")));
+            else
+                explanation = i18n("non object JSON value");
+
+            m_isLoadingFromFile = false;
+            showInvalidNotebookSchemeError(i18n("found incorrect data (%1) that is not Jupyter cell", explanation));
+            return false;
+        }
+
+        const QJsonObject& cell = iter->toObject();
+        QString cellType = Cantor::JupyterUtils::getCellType(cell);
+
+        if (cellType == QLatin1String("code"))
+        {
+            if (LatexEntry::isConvertableToLatexEntry(cell))
+            {
+                entry = appendEntry(LatexEntry::Type, false);
+                entry->setContentFromJupyter(cell);
+                entry->evaluate(WorksheetEntry::InternalEvaluation);
+            }
+            else
+            {
+                entry = appendEntry(CommandEntry::Type, false);
+                entry->setContentFromJupyter(cell);
+            }
+        }
+        else if (cellType == QLatin1String("markdown"))
+        {
+            if (TextEntry::isConvertableToTextEntry(cell))
+            {
+                entry = appendEntry(TextEntry::Type, false);
+                entry->setContentFromJupyter(cell);
+            }
+            else
+            {
+                entry = appendEntry(MarkdownEntry::Type, false);
+                entry->setContentFromJupyter(cell);
+                entry->evaluate(WorksheetEntry::InternalEvaluation);
+            }
+        }
+        else if (cellType == QLatin1String("raw"))
+        {
+            if (PageBreakEntry::isConvertableToPageBreakEntry(cell))
+                entry = appendEntry(PageBreakEntry::Type, false);
+            else
+                entry = appendEntry(TextEntry::Type, false);
+            entry->setContentFromJupyter(cell);
+        }
+
+        if (m_readOnly && entry)
+        {
+            entry->setAcceptHoverEvents(false);
+            entry = nullptr;
+        }
+    }
+
+    if (m_readOnly)
+        clearFocus();
+
+    m_isLoadingFromFile = false;
+    updateLayout();
+
+    enableHighlighting( m_highlighter!=nullptr || Settings::highlightDefault() );
+
+    emit loaded();
+    return true;
+}
+
+void Worksheet::showInvalidNotebookSchemeError(QString additionalInfo)
+{
+    if (additionalInfo.isEmpty())
+        KMessageBox::error(worksheetView(), i18n("The file is not valid Jupyter notebook"), i18n("Open File"));
+    else
+        KMessageBox::error(worksheetView(), i18n("Invalid Jupyter notebook scheme: %1", additionalInfo), i18n("Open File"));
+}
+
+
 void Worksheet::gotResult(Cantor::Expression* expr)
 {
     if(expr==nullptr)
@@ -1208,11 +1512,16 @@ void Worksheet::removeCurrentEntry()
     entry->startRemoving();
 }
 
-EpsRenderer* Worksheet::epsRenderer()
+Cantor::Renderer* Worksheet::renderer()
 {
     return &m_epsRenderer;
 }
 
+MathRenderer* Worksheet::mathRenderer()
+{
+    return &m_mathRenderer;
+}
+
 QMenu* Worksheet::createContextMenu()
 {
     QMenu *menu = new QMenu(worksheetView());
@@ -1240,9 +1549,21 @@ void Worksheet::populateMenu(QMenu *menu, QPointF pos)
     menu->addSeparator();
 
     if (entry) {
+        QMenu* convertTo = new QMenu(menu);
         QMenu* insert = new QMenu(menu);
         QMenu* insertBefore = new QMenu(menu);
 
+        convertTo->addAction(QIcon::fromTheme(QLatin1String("run-build")), i18n("Command Entry"), entry, &WorksheetEntry::convertToCommandEntry);
+        convertTo->addAction(QIcon::fromTheme(QLatin1String("draw-text")), i18n("Text Entry"), entry, &WorksheetEntry::convertToTextEntry);
+#ifdef Discount_FOUND
+        convertTo->addAction(QIcon::fromTheme(QLatin1String("text-x-markdown")), i18n("Markdown Entry"), entry, &WorksheetEntry::convertToMarkdownEntry);
+#endif
+#ifdef WITH_EPS
+        convertTo->addAction(QIcon::fromTheme(QLatin1String("text-x-tex")), i18n("LaTeX Entry"), entry, &WorksheetEntry::convertToLatexEntry);
+#endif
+        convertTo->addAction(QIcon::fromTheme(QLatin1String("image-x-generic")), i18n("Image"), entry, &WorksheetEntry::convertToImageEntry);
+        convertTo->addAction(QIcon::fromTheme(QLatin1String("go-next-view-page")), i18n("Page Break"), entry, &WorksheetEntry::converToPageBreakEntry);
+
         insert->addAction(QIcon::fromTheme(QLatin1String("run-build")), i18n("Command Entry"), entry, SLOT(insertCommandEntry()));
         insert->addAction(QIcon::fromTheme(QLatin1String("draw-text")), i18n("Text Entry"), entry, SLOT(insertTextEntry()));
 #ifdef Discount_FOUND
@@ -1265,10 +1586,14 @@ void Worksheet::populateMenu(QMenu *menu, QPointF pos)
         insertBefore->addAction(QIcon::fromTheme(QLatin1String("image-x-generic")), i18n("Image"), entry, SLOT(insertImageEntryBefore()));
         insertBefore->addAction(QIcon::fromTheme(QLatin1String("go-next-view-page")), i18n("Page Break"), entry, SLOT(insertPageBreakEntryBefore()));
 
+        convertTo->setTitle(i18n("Convert Entry To"));
+        convertTo->setIcon(QIcon::fromTheme(QLatin1String("gtk-convert")));
         insert->setTitle(i18n("Insert Entry After"));
         insert->setIcon(QIcon::fromTheme(QLatin1String("edit-table-insert-row-below")));
         insertBefore->setTitle(i18n("Insert Entry Before"));
         insertBefore->setIcon(QIcon::fromTheme(QLatin1String("edit-table-insert-row-above")));
+
+        menu->addMenu(convertTo);
         menu->addMenu(insert);
         menu->addMenu(insertBefore);
     } else {
@@ -1957,3 +2282,76 @@ void Worksheet::drawEntryCursor()
         m_entryCursorItem->show();
     }
 }
+
+void Worksheet::setType(Worksheet::Type type)
+{
+    m_type = type;
+}
+
+Worksheet::Type Worksheet::type() const
+{
+    return m_type;
+}
+
+void Worksheet::changeEntryType(WorksheetEntry* target, int newType)
+{
+    if (target && target->type() != newType)
+    {
+        bool animation_state = m_animationsEnabled;
+        m_animationsEnabled = false;
+
+        QString content;
+
+        switch(target->type())
+        {
+            case CommandEntry::Type:
+                content = static_cast<CommandEntry*>(target)->command();
+                break;
+
+            case MarkdownEntry::Type:
+                content = static_cast<MarkdownEntry*>(target)->plainText();
+                break;
+
+            case TextEntry::Type:
+                content = static_cast<TextEntry*>(target)->text();
+                break;
+
+            case LatexEntry::Type:
+                content = static_cast<LatexEntry*>(target)->plain();
+
+        }
+
+        WorksheetEntry* newEntry = WorksheetEntry::create(newType, this);
+        newEntry->setContent(content);
+
+        if (newEntry)
+        {
+            WorksheetEntry* tmp = target;
+
+            newEntry->setPrevious(tmp->previous());
+            newEntry->setNext(tmp->next());
+
+            tmp->setPrevious(nullptr);
+            tmp->setNext(nullptr);
+            tmp->clearFocus();
+            tmp->forceRemove();
+
+            if (newEntry->previous())
+                newEntry->previous()->setNext(newEntry);
+            else
+                setFirstEntry(newEntry);
+
+            if (newEntry->next())
+                newEntry->next()->setPrevious(newEntry);
+            else
+                setLastEntry(newEntry);
+
+            updateLayout();
+            makeVisible(newEntry);
+            focusEntry(newEntry);
+            setModified();
+            newEntry->focusEntry();
+        }
+        m_animationsEnabled = animation_state;
+    }
+}
