@@ -32,7 +32,8 @@
 #include <QMenu>
 
 #include "worksheetview.h"
-#include "epsrenderer.h"
+#include "lib/renderer.h"
+#include "mathrender.h"
 #include "worksheetcursor.h"
 
 namespace Cantor {
@@ -57,7 +58,12 @@ class Worksheet : public QGraphicsScene
 {
   Q_OBJECT
   public:
-    Worksheet(Cantor::Backend* backend, QWidget* parent);
+    enum Type {
+      CantorWorksheet,
+      JupyterNotebook
+    };
+
+    Worksheet(Cantor::Backend* backend, QWidget* parent, bool useDeafultWorksheetParameters = true);
     ~Worksheet() override;
 
     Cantor::Session* session();
@@ -68,11 +74,10 @@ class Worksheet : public QGraphicsScene
     bool isReadOnly();
     bool showExpressionIds();
     bool animationsEnabled();
+    bool embeddedMathEnabled();
 
     bool isPrinting();
 
-    void setViewSize(qreal w, qreal h, qreal s, bool forceUpdate = false);
-
     WorksheetView* worksheetView();
 
     void makeVisible(WorksheetEntry*);
@@ -85,7 +90,8 @@ class Worksheet : public QGraphicsScene
     void createActions(KActionCollection*);
     QMenu* createContextMenu();
     void populateMenu(QMenu* menu, QPointF pos);
-    EpsRenderer* epsRenderer();
+    Cantor::Renderer* renderer();
+    MathRenderer* mathRenderer();
     bool isEmpty();
     bool isLoadingFromFile();
 
@@ -101,12 +107,26 @@ class Worksheet : public QGraphicsScene
     // For WorksheetEntry::startDrag
     void resetEntryCursor();
 
-    void addProtrusion(qreal width);
-    void updateProtrusion(qreal oldWidth, qreal newWidth);
-    void removeProtrusion(qreal width);
+    /**
+     * How it works:
+     * There are two information streams
+     * 1. WorksheetView -> Worksheet -> subelemenets (ex. entries) about view width
+     *   View width used by some sub elements for better visual appearance (for example, entries with text often are fitted to width of view).
+     * 2. Subelements -> Worksheet
+     *   Sub elements notify Worksheet about their needed widths and worksheet, used this information, set proper scene size.
+     */
+    /// First information stream
+    void setViewSize(qreal w, qreal h, qreal s, bool forceUpdate = false);
+
+    /// Second information stream
+    void setRequestedWidth(QGraphicsObject* object, qreal width);
+    void removeRequestedWidth(QGraphicsObject* object);
 
     bool isShortcut(const QKeySequence&);
 
+    void setType(Worksheet::Type type);
+    Worksheet::Type type() const;
+
     // richtext
     struct RichTextInfo {
         bool bold;
@@ -162,6 +182,7 @@ class Worksheet : public QGraphicsScene
     void enableCompletion(bool);
     void enableExpressionNumbering(bool);
     void enableAnimations(bool);
+    void enableEmbeddedMath(bool);
 
     QDomDocument toXML(KZip* archive = nullptr);
 
@@ -207,6 +228,8 @@ class Worksheet : public QGraphicsScene
     void setFontFamily(const QString&);
     void setFontSize(int size);
 
+    void changeEntryType(WorksheetEntry* target, int newType);
+
   Q_SIGNALS:
     void modified();
     void loaded();
@@ -233,6 +256,8 @@ class Worksheet : public QGraphicsScene
 
     void keyPressEvent(QKeyEvent*) override;
 
+    QJsonDocument toJupyterJson();
+
   private Q_SLOTS:
     void showCompletion();
     //void checkEntriesForSanity();
@@ -251,6 +276,10 @@ class Worksheet : public QGraphicsScene
     void addEntryFromEntryCursor();
     void drawEntryCursor();
     int entryCount();
+    bool loadCantorWorksheet(const KZip& archive);
+    bool loadJupyterNotebook(const QJsonDocument& doc);
+    void showInvalidNotebookSchemeError(QString additionalInfo = QString());
+    void initSession(Cantor::Backend*);
 
   private:
     static const double LeftMargin;
@@ -260,7 +289,8 @@ class Worksheet : public QGraphicsScene
     static const double EntryCursorWidth;
     Cantor::Session *m_session;
     QSyntaxHighlighter* m_highlighter;
-    EpsRenderer m_epsRenderer;
+    Cantor::Renderer m_epsRenderer;
+    MathRenderer m_mathRenderer;
     WorksheetEntry* m_firstEntry;
     WorksheetEntry* m_lastEntry;
     WorksheetEntry* m_dragEntry;
@@ -272,9 +302,9 @@ class Worksheet : public QGraphicsScene
     WorksheetTextItem* m_lastFocusedTextItem;
     QTimer* m_dragScrollTimer;
 
-    double m_viewWidth;
-    double m_protrusion;
-    QMap<qreal, int> m_itemProtrusions;
+    qreal m_viewWidth;
+    QMap<QGraphicsObject*, qreal> m_itemWidths;
+    qreal m_maxWidth;
 
     QMap<QKeySequence, QAction*> m_shortcuts;
 
@@ -290,15 +320,21 @@ class Worksheet : public QGraphicsScene
     KToggleAction* m_alignRightAction;
     KToggleAction* m_alignJustifyAction;
 
-    bool m_completionEnabled;
-    bool m_showExpressionIds;
-    bool m_animationsEnabled;
-    bool m_loginDone;
-    bool m_isPrinting;
-    bool m_isLoadingFromFile;
-    bool m_readOnly;
+    bool m_useDefaultWorksheetParameters{true};
+
+    bool m_completionEnabled{false};
+    bool m_embeddedMathEnabled{false};
+    bool m_showExpressionIds{false};
+    bool m_animationsEnabled{false};
+
+    bool m_isPrinting{false};
+    bool m_isLoadingFromFile{false};
+    bool m_readOnly{false};
+
+    Type m_type = CantorWorksheet;
 
     QString m_backendName;
+    QJsonObject* m_jupyterMetadata{nullptr};
 };
 
 #endif // WORKSHEET_H
