@@ -21,8 +21,9 @@
 #include "worksheettextitem.h"
 #include "worksheet.h"
 #include "worksheetentry.h"
-#include "epsrenderer.h"
+#include "lib/renderer.h"
 #include "worksheetcursor.h"
+#include "extended_document.h"
 
 #include <QApplication>
 #include <QClipboard>
@@ -45,6 +46,7 @@
 WorksheetTextItem::WorksheetTextItem(QGraphicsObject* parent, Qt::TextInteractionFlags ti)
     : QGraphicsTextItem(parent)
 {
+    setDocument(new ExtendedDocument(this));
     setTextInteractionFlags(ti);
     if (ti & Qt::TextEditable) {
         setCursor(Qt::IBeamCursor);
@@ -56,7 +58,6 @@ WorksheetTextItem::WorksheetTextItem(QGraphicsObject* parent, Qt::TextInteractio
     m_itemDragable = false;
     m_richTextEnabled = false;
     m_size = document()->size();;
-    m_maxWidth = -1;
     setAcceptDrops(true);
     setFont(QFontDatabase::systemFont(QFontDatabase::FixedFont));
     connect(document(), SIGNAL(contentsChanged()), this, SLOT(testSize()));
@@ -74,8 +75,9 @@ WorksheetTextItem::~WorksheetTextItem()
 {
     if (worksheet() && this == worksheet()->lastFocusedTextItem())
         worksheet()->updateFocusedTextItem(nullptr);
-    if (worksheet() && m_maxWidth > 0 && width() > m_maxWidth)
-        worksheet()->removeProtrusion(width() - m_maxWidth);
+
+    if (worksheet())
+        worksheet()->removeRequestedWidth(this);
 }
 
 int WorksheetTextItem::type() const
@@ -100,19 +102,11 @@ void WorksheetTextItem::testSize()
 
     qreal w = document()->size().width();
     if (w != m_size.width()) {
-        if (m_maxWidth > 0) {
-            qreal oldDiff = m_size.width() - m_maxWidth;
-            qreal newDiff = w - m_maxWidth;
-            if (w > m_maxWidth) {
-                if (m_size.width() > m_maxWidth)
-                    worksheet()->updateProtrusion(oldDiff, newDiff);
-                else
-                    worksheet()->addProtrusion(newDiff);
-            } else if (m_size.width() > m_maxWidth) {
-                worksheet()->removeProtrusion(oldDiff);
-            }
-        }
         m_size.setWidth(w);
+        emit sizeChanged();
+
+        qreal newWidth = scenePos().x() + m_size.width() - 10;
+        worksheet()->setRequestedWidth(this, newWidth);
     }
 }
 
@@ -123,24 +117,14 @@ qreal WorksheetTextItem::setGeometry(qreal x, qreal y, qreal w, bool centered)
     else
         setPos(x,y);
 
-    qreal oldDiff = 0;
-    if (m_maxWidth > 0 && width() > m_maxWidth)
-        oldDiff = width() - m_maxWidth;
-    m_maxWidth = w;
+    // Strange: if I use the same logic as for ImageItem (with scenePos.x() + width)
+    // Cantor always have scrollbar for a few pixels
+    // So I always subtract the few pixels
     setTextWidth(w);
     m_size = document()->size();
 
-    if (oldDiff) {
-        if (m_size.width() > m_maxWidth) {
-            qreal newDiff = m_size.width() - m_maxWidth;
-            worksheet()->updateProtrusion(oldDiff, newDiff);
-        } else {
-            worksheet()->removeProtrusion(oldDiff);
-        }
-    } else if (m_size.width() > m_maxWidth) {
-        qreal newDiff = m_size.width() - m_maxWidth;
-        worksheet()->addProtrusion(newDiff);
-    }
+    qreal newWidth = scenePos().x() + m_size.width() - 10;
+    worksheet()->setRequestedWidth(this, newWidth);
 
     return m_size.height();
 }
@@ -270,13 +254,13 @@ QString WorksheetTextItem::resolveImages(const QTextCursor& cursor)
          cursor2 = document()->find(repl, cursor1)) {
         cursor1.setPosition(cursor2.selectionStart(), QTextCursor::KeepAnchor);
         result += cursor1.selectedText();
-        QVariant var = cursor2.charFormat().property(EpsRenderer::Delimiter);
+        QVariant var = cursor2.charFormat().property(Cantor::Renderer::Delimiter);
         QString delim;
         if (var.isValid())
             delim = var.value<QString>();
         else
             delim = QLatin1String("");
-        result += delim + cursor2.charFormat().property(EpsRenderer::Code).value<QString>() + delim;
+        result += delim + cursor2.charFormat().property(Cantor::Renderer::Code).value<QString>() + delim;
         cursor1.setPosition(cursor2.selectionEnd());
     }
 
@@ -499,7 +483,7 @@ bool WorksheetTextItem::sceneEvent(QEvent *event)
         // QGraphicsTextItem's TabChangesFocus feature prevents calls to
         // keyPressEvent for Tab, even when it's turned off. So we got to catch
         // that here.
-        QKeyEvent* kev = dynamic_cast<QKeyEvent*>(event);
+        QKeyEvent* kev = static_cast<QKeyEvent*>(event);
         if (kev->key() == Qt::Key_Tab && kev->modifiers() == Qt::NoModifier) {
             emit tabPressed();
             return true;
@@ -510,7 +494,7 @@ bool WorksheetTextItem::sceneEvent(QEvent *event)
             return true;
         }
     } else if (event->type() == QEvent::ShortcutOverride) {
-        QKeyEvent* kev = dynamic_cast<QKeyEvent*>(event);
+        QKeyEvent* kev = static_cast<QKeyEvent*>(event);
         QKeySequence seq(kev->key() + kev->modifiers());
         if (worksheet()->isShortcut(seq)) {
             qDebug() << "ShortcutOverride" << kev->key() << kev->modifiers();
@@ -922,5 +906,5 @@ void WorksheetTextItem::allowEditing()
 
 void WorksheetTextItem::denyEditing()
 {
-    setTextInteractionFlags(Qt::TextSelectableByMouse | Qt::TextSelectableByKeyboard);
+    setTextInteractionFlags(Qt::TextBrowserInteraction | Qt::TextSelectableByKeyboard);
 }
